/* GNU FaXile - fax image viewer
 *
 * Copyright (C) 2000 Wolfgang Sourdeau
 *
 * Author: Wolfgang Sourdeau <wolfgang@ultim.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef ENABLE_NLS
#include <locale.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gtk/gtk.h>

#ifdef __WIN32__
#include <windows.h>
#endif

#include "tiffimages.h"
#include "viewer.h"
#include "print.h"

#ifdef __unix__
/* Dumb UNIX printing */

#define MAX_PLIST 8

static inline unsigned char reverse_bits (unsigned char byte)
{
	unsigned char result;

	result = 0;

        if (byte & 128) result |= 1;
        if (byte & 64)  result |= 2;
        if (byte & 32)  result |= 4;
        if (byte & 16)  result |= 8;
        if (byte & 8)   result |= 16;
        if (byte & 4)   result |= 32;
        if (byte & 2)   result |= 64;
        if (byte & 1)   result |= 128;

	return result;
}

static int compare_printers(PList *p1, PList *p2)
{
	return (strcasecmp(p1->name, p2->name));
}

static PList *get_printers(void)
{
	int i, plist_count;
	char defname[17];
	PList *plist;
#if defined(LPC_COMMAND) || defined(LPSTAT_COMMAND)
	FILE *pfile;
	char line[129], name[17];
#endif	
	
	defname[0] = '\0';
	
	plist = g_malloc0 (sizeof (PList) * MAX_PLIST);
	plist_count = 0;
	
#ifdef LPC_COMMAND
	if ((pfile = popen(LPC_COMMAND " status all < /dev/null", "r"))
	    != NULL) {
		while (fgets(line, sizeof (line), pfile) != NULL
		       && plist_count < MAX_PLIST)
			if (strchr(line, ':') != NULL
			    && line[0] != ' '
			    && line[0] != '\t'
			    && strncmp(line,
				       "Press RETURN to continue"
				       , 24)) {
				*strchr(line, ':') = '\0';
				strcpy(plist[plist_count].name, line);
				plist[plist_count].active = TRUE;
				plist_count++;
			}
		
		pclose(pfile);
	}
#endif /* LPC_COMMAND */
	
#ifdef LPSTAT_COMMAND
	if (!plist_count
	    && (pfile = popen(LPSTAT_COMMAND " -d -p", "r")) != NULL) {
		while (fgets(line, sizeof (line), pfile) != NULL &&
		       plist_count < MAX_PLIST) {
			if (sscanf(line, "printer %s", name) == 1) {
				strcpy(plist[plist_count].name, name);
				plist[plist_count].active = 1;
				plist_count++;
			} else
				sscanf(line, 
				       "system default destination:"
				       " %s",
				       defname);
		}
		
		pclose(pfile);
	}
#endif /* LPSTAT_COMMAND */
	
	if (plist_count > 1)
		qsort(plist, plist_count, sizeof (PList),
		      (__compar_fn_t) compare_printers);
	else if (plist_count == 0) {
		g_free (plist);
		plist = NULL;
	}

	return plist;
}

gint activate_printer_cb (GtkWidget *widget, PrintData *print_data)
{
	print_data->printer =
		gtk_object_get_data (GTK_OBJECT (widget),
				     "printer_name");

	return FALSE;
}

GtkWidget *make_printer_menu (PrintData *print_data)
{
	GtkWidget *menu, *menuitem;
	PList *plist, *cur_printer;

	plist = get_printers ();

	if (plist) {
		menu = gtk_menu_new ();
		cur_printer = plist;

		if (cur_printer->active)
			print_data->printer = cur_printer->name;

		while (cur_printer->active) {
			menuitem = gtk_menu_item_new_with_label
				(cur_printer->name);
			gtk_menu_append (GTK_MENU (menu), menuitem);
			gtk_object_set_data (GTK_OBJECT (menuitem),
					     "printer_name",
					     cur_printer->name);
			gtk_signal_connect (GTK_OBJECT (menuitem),
					    "activate",
					    GTK_SIGNAL_FUNC
					    (activate_printer_cb),
					    print_data);
			gtk_widget_show (menuitem);
			cur_printer++;
		}

		gtk_object_set_data_full (GTK_OBJECT (menu),
					  "printer_list",
					  plist,
					  g_free);
	} else
		menu = NULL;

	return menu;
}

gint toggle_entries_cb (GtkWidget *widget, PDlgWidgets *widgets)
{
	GtkWidget *menu, *own_text, *other_text;

	if (widget == widgets->printer_rb) {
		gtk_widget_set_sensitive (widgets->printer_opt_menu, TRUE);
		gtk_widget_set_sensitive (widgets->printer_cmd_en, TRUE);
		gtk_widget_set_sensitive (widgets->output_file_en, FALSE);
	} else {
		gtk_widget_set_sensitive (widgets->printer_opt_menu, FALSE);
		gtk_widget_set_sensitive (widgets->printer_cmd_en, FALSE);
		gtk_widget_set_sensitive (widgets->output_file_en, TRUE);
	}

	return FALSE;
}

gint toggle_page_sel_entries_cb (GtkWidget *widget,
				 PDlgWidgets *widgets)
{
	if (widget == widgets->from_to_rb) {
		gtk_widget_set_sensitive (widgets->from_sp_but,
					  TRUE);
		gtk_widget_set_sensitive (widgets->to_sp_but,
					  TRUE);
	} else {
		gtk_widget_set_sensitive (widgets->from_sp_but,
					  FALSE);
		gtk_widget_set_sensitive (widgets->to_sp_but,
					  FALSE);
	}

	return FALSE;
}

GtkWidget *make_output_frame (PrintData *print_data)
{
	GtkWidget *output_frame, *table, *output_separator;
	GtkWidget *printer_menu;
	GtkWidget *printer_cmd_lbl, *output_file_lbl;
	GSList *rb_group;
	gchar *output_file_txt;

	output_frame = gtk_frame_new (_("Output to"));
	gtk_container_set_border_width (GTK_CONTAINER (output_frame), 5);
	table = gtk_table_new (5, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 3);
	gtk_table_set_row_spacings (GTK_TABLE (table), 3);
	gtk_table_set_col_spacings (GTK_TABLE (table), 3);
	output_separator = gtk_hseparator_new ();

	print_data->widgets.printer_rb =
		gtk_radio_button_new_with_label (NULL, _("Printer"));

	rb_group = gtk_radio_button_group
		(GTK_RADIO_BUTTON(print_data->widgets.printer_rb));
	print_data->widgets.file_rb =
		gtk_radio_button_new_with_label (rb_group, _("File"));

	gtk_signal_connect (GTK_OBJECT
			    (print_data->widgets.printer_rb),
			    "clicked",
			    GTK_SIGNAL_FUNC (toggle_entries_cb),
			    &(print_data->widgets));			     
	gtk_signal_connect (GTK_OBJECT (print_data->widgets.file_rb),
			    "clicked",
			    GTK_SIGNAL_FUNC (toggle_entries_cb),
			    &(print_data->widgets));			     

	gtk_widget_show (print_data->widgets.printer_rb);
	gtk_widget_show (print_data->widgets.file_rb);

	printer_cmd_lbl = gtk_label_new (_("Command"));
	gtk_misc_set_alignment (GTK_MISC (printer_cmd_lbl), 0.0, 0.5);
	output_file_lbl = gtk_label_new (_("File"));
	gtk_misc_set_alignment (GTK_MISC (output_file_lbl), 0.0, 0.5);

	print_data->widgets.printer_cmd_en = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY
			    (print_data->widgets.printer_cmd_en),
			    "lpr");
	gtk_widget_set_usize (print_data->widgets.printer_cmd_en,
			      80, 20);
	print_data->widgets.output_file_en = gtk_entry_new ();
	output_file_txt = _("fax_output.ps");
	gtk_entry_set_text (GTK_ENTRY
			    (print_data->widgets.output_file_en),
			    output_file_txt);
	gtk_editable_set_position (GTK_EDITABLE
				   (print_data->widgets.output_file_en),
				   strlen (output_file_txt));
	gtk_widget_set_sensitive (print_data->widgets.output_file_en,
				  FALSE);
	gtk_widget_set_usize (print_data->widgets.output_file_en,
			      80, 20); 

	print_data->widgets.printer_opt_menu = gtk_option_menu_new ();
	printer_menu = make_printer_menu (print_data);

	gtk_widget_set_usize (print_data->widgets.printer_opt_menu, 80, 25);

	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_data->widgets.printer_rb,
				   0, 1, 0, 1);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_data->widgets.printer_opt_menu,
				   1, 2, 0, 1);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   printer_cmd_lbl,
				   0, 1, 1, 2);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_data->widgets.printer_cmd_en,
				   1, 2, 1, 2);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   output_separator,
				   0, 2, 2, 3);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_data->widgets.file_rb,
				   0, 1, 3, 4);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   output_file_lbl,
				   0, 1, 4, 5);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_data->widgets.output_file_en,
				   1, 2, 4, 5);

	gtk_widget_show (table);
	gtk_container_add (GTK_CONTAINER (output_frame), table);

	gtk_widget_show (output_separator);
	gtk_widget_show (print_data->widgets.printer_opt_menu);
	gtk_widget_show (print_data->widgets.printer_cmd_en);
	gtk_widget_show (printer_cmd_lbl);
	gtk_widget_show (print_data->widgets.output_file_en);
	gtk_widget_show (output_file_lbl);

	if (printer_menu) {
		gtk_widget_set_usize (printer_menu, 79, -2);
		gtk_option_menu_set_menu
			(GTK_OPTION_MENU
			 (print_data->widgets.printer_opt_menu),
			 printer_menu);
		gtk_widget_show (printer_menu);
	} else {
		gtk_button_clicked (GTK_BUTTON (print_data->widgets.file_rb));
		gtk_widget_set_sensitive
			(print_data->widgets.printer_opt_menu, FALSE);
		gtk_widget_set_sensitive
			(print_data->widgets.printer_rb, FALSE);
	}

	return output_frame;
}

GtkWidget *make_page_frame (PrintData *print_data)
{
	GtkWidget *page_frame, *page_table;
	GtkWidget *to_label;
	GtkObject *from_adj, *to_adj;
	gfloat max_pages;
	GSList *rb_group;

	page_frame = gtk_frame_new (_("Page selection"));
	gtk_container_set_border_width (GTK_CONTAINER (page_frame),
					5);

	page_table = gtk_table_new (3, 4, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (page_table), 3);
	gtk_container_add (GTK_CONTAINER (page_frame), page_table);
	gtk_widget_show (page_table);

	print_data->widgets.all_pages_rb = 
		gtk_radio_button_new_with_label (NULL,
						 _("All pages"));
	rb_group = gtk_radio_button_group
		(GTK_RADIO_BUTTON (print_data->widgets.all_pages_rb));
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   print_data->widgets.all_pages_rb,
				   0, 4, 0, 1);
	gtk_signal_connect (GTK_OBJECT
			    (print_data->widgets.all_pages_rb),
			    "clicked",
			    GTK_SIGNAL_FUNC
			    (toggle_page_sel_entries_cb),
			    &(print_data->widgets));
	gtk_widget_show (print_data->widgets.all_pages_rb);

	print_data->widgets.from_to_rb =
		gtk_radio_button_new_with_label (rb_group, _("From"));
	rb_group = gtk_radio_button_group (GTK_RADIO_BUTTON
					   (print_data->widgets.from_to_rb));
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   print_data->widgets.from_to_rb,
				   0, 1, 1, 2);
	gtk_signal_connect (GTK_OBJECT
			    (print_data->widgets.from_to_rb),
			    "clicked",
			    GTK_SIGNAL_FUNC
			    (toggle_page_sel_entries_cb),
			    &(print_data->widgets));
	gtk_widget_show (print_data->widgets.from_to_rb);

	max_pages = (gfloat) (print_data->document->nbr_images + 1);
	from_adj = gtk_adjustment_new (1.0,
				       1.0, max_pages,
				       1.0, 0.0, 0.0);

	print_data->widgets.from_sp_but = gtk_spin_button_new
		(GTK_ADJUSTMENT (from_adj), 1.0, 0);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON
				     (print_data->widgets.from_sp_but),
				     TRUE);
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   print_data->widgets.from_sp_but,
				   1, 2, 1, 2);
	gtk_widget_show (print_data->widgets.from_sp_but);

	to_label = gtk_label_new (_("to"));
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   to_label,
				   2, 3, 1, 2);
	gtk_widget_show (to_label);

	to_adj = gtk_adjustment_new (max_pages,
				     1.0, max_pages,
				     1.0, 0.0, 0.0);

	print_data->widgets.to_sp_but =
		gtk_spin_button_new (GTK_ADJUSTMENT (to_adj), 1.0, 0);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON
				     (print_data->widgets.to_sp_but),
				     TRUE);
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   print_data->widgets.to_sp_but,
				   3, 4, 1, 2);
	gtk_widget_show (print_data->widgets.to_sp_but);

	print_data->widgets.cur_page_rb =
		gtk_radio_button_new_with_label
		(rb_group, _("Current page only"));
	gtk_table_attach_defaults (GTK_TABLE (page_table),
				   print_data->widgets.cur_page_rb,
				   0, 4, 2, 3);
	gtk_signal_connect (GTK_OBJECT
			    (print_data->widgets.cur_page_rb),
			    "clicked",
			    GTK_SIGNAL_FUNC
			    (toggle_page_sel_entries_cb),
			    &(print_data->widgets));
	gtk_widget_show (print_data->widgets.cur_page_rb);

	gtk_widget_set_sensitive (print_data->widgets.from_sp_but,
				  FALSE);
	gtk_widget_set_sensitive (print_data->widgets.to_sp_but,
				  FALSE);

	return page_frame;
}

FILE *create_temp_print_file (gchar **outfilename)
{
	gint counter;
	FILE *file_stream;
	char *file_name, *template;

	file_stream = NULL;
	counter = 0;

	template = g_strdup ("/tmp/gfo-XXXXXX");

	while (!file_stream && counter < 64) {
		file_name = mktemp (template);
		file_stream = fopen (template, "wx");
		if (file_stream)
			*outfilename = template;
		counter++;
	}

	return file_stream;
}

void print_image (FILE *print_file, FaxImage *fax_image)
{
	unsigned int full_length, counter;

	full_length = (int) ((float) fax_image->width
			     * (float) fax_image->height
			     / 8.0);

	for (counter = 0; counter < full_length; counter++) {
		fprintf(print_file, "%.2x", 
			reverse_bits (~*(fax_image->image
					 + counter)));
		
		if ((counter + 1) % 40 == 0)
			fprintf(print_file, "\n");
	}

	if ((counter + 1) % 40 != 0)
		fprintf(print_file, "\n");
}

void output_document (OutputData *output_data)
{
	gint cur_page_nbr;
	FaxImage *cur_page;
	gchar *time_string;
	time_t cur_time;

#ifdef ENABLE_NLS
/* If we don't do this, our floating-point numbers might become
 * floating-comma numbers... */
	setlocale (LC_NUMERIC, "C");
#endif

	cur_page_nbr = output_data->from_page;
	time (&cur_time);
	time_string = ctime (&cur_time);

	fprintf(output_data->output_stream,
		"%%!PS-Adobe-3.0\n"
		"%%%%Creator: GNU FaXile - Viewer\n"
		"%%%%Title: Fax Output\n"
		"%%%%CreationDate: %s"
		"%%%%DocumentData: Clean7Bit\n"
		"%%%%Origin: 0 0\n"
		"%%%%BoundingBox: 0 0 612 792\n"
		"%%%%LanguageLevel: 1\n"
		"%%%%Pages: (atend)\n"
		"%%%%EndComments\n"
		"%%%%BeginSetup\n"
		"%%%%EndSetup\n",
		time_string);

	while (!(cur_page_nbr > output_data->to_page)) {
		cur_page = get_fax_image (output_data->document, cur_page_nbr - 1);

		fprintf(output_data->output_stream,
			"%%%%Page: %d %d\n", cur_page_nbr, cur_page_nbr);
		fprintf(output_data->output_stream,
			"gsave\n"
			"100 dict begin\n"
			"%f %f scale\n",
			((float) cur_page->width
			 * 72.0
			 / (float) output_data->document->xres),
			((float) cur_page->height
			 * 72.0
			 / (float) output_data->document->yres));
		fprintf(output_data->output_stream,
			"%%ImageData: %d %d 1 1 0 1 2 \"image\"\n",
			cur_page->width, cur_page->height);
		fprintf(output_data->output_stream,
			"/scanLine %d string def\n",
			(int) (cur_page->width / 8));
		fprintf(output_data->output_stream,
			"%d %d 1\n", cur_page->width, cur_page->height);
		fprintf(output_data->output_stream,
			"[%d 0 0 -%d 0 %d]\n", cur_page->width,
			cur_page->height, cur_page->height);
		fprintf(output_data->output_stream,
			"{currentfile scanLine readhexstring pop} bind\n"
			"image\n");

		load_tiff_image (output_data->document, cur_page);
		print_image (output_data->output_stream, cur_page);
		unload_tiff_image (cur_page);

		fprintf(output_data->output_stream,
			"end\n"
			"grestore\n"
			"showpage\n");

		cur_page_nbr++;
	}

	fprintf (output_data->output_stream,
		 "%%%%Trailer\n"
		 "%%%%Pages: %d\n"
		 "%%%%EOF\n", (cur_page_nbr - output_data->from_page));

#ifdef ENABLE_NLS
	setlocale (LC_NUMERIC, "");
#endif
}

gint print_to_file_anyway_cb (GtkWidget *yes_button, OutputData
			      *output_data)
{
	gtk_widget_destroy (output_data->err_dlg);

	while (gtk_events_pending ())
		gtk_main_iteration ();

	output_data->output_stream =
		fopen (output_data->out_filename, "w+");
	output_document (output_data);
	fclose (output_data->output_stream);
	g_free (output_data->out_filename);
	g_free (output_data);

	gtk_widget_destroy (output_data->print_dlg);

	return FALSE;
}

gint no_overwrite_cb (GtkWidget *no_button, OutputData *output_data)
{
	gtk_widget_destroy (output_data->err_dlg);
	gtk_window_set_modal (GTK_WINDOW (output_data->print_dlg),
			      TRUE);
	g_free (output_data->out_filename);
	g_free (output_data);

	return FALSE;
}

void file_exists_dlg (OutputData *output_data,
		      GtkWidget *print_dlg)
{
	GtkWidget *err_win, *vbox, *hbox, *msg_lbl, *yes_but, *no_but;
	gchar *message;

	message = g_strdup_printf (_("%s already exists.\n"
				     "Do you want to overwrite it?"),
				   output_data->out_filename);

	err_win = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_policy (GTK_WINDOW (err_win), FALSE, FALSE,
			       TRUE);
	gtk_window_set_title (GTK_WINDOW (err_win), _("Please answer..."));
	gtk_window_set_modal (GTK_WINDOW (err_win), TRUE);
	gtk_window_set_modal (GTK_WINDOW (print_dlg), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (err_win), 10);

	vbox = gtk_vbox_new (FALSE, 1);
	gtk_container_add (GTK_CONTAINER (err_win), vbox);
	gtk_widget_show (vbox);
	
	msg_lbl = gtk_label_new (message);
	g_free (message);
	gtk_box_pack_start (GTK_BOX (vbox), msg_lbl, TRUE, FALSE, 3);
	gtk_label_set_justify (GTK_LABEL (msg_lbl),
			       GTK_JUSTIFY_LEFT);
	gtk_widget_show (msg_lbl);

	hbox = gtk_hbox_new (FALSE, 3);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, FALSE, 0);
	gtk_widget_show (hbox);

	output_data->err_dlg = err_win;
	output_data->print_dlg = print_dlg;

	yes_but = gtk_button_new_with_label (_("Yes, please do"));
	gtk_signal_connect (GTK_OBJECT (yes_but), "clicked",
			    GTK_SIGNAL_FUNC (print_to_file_anyway_cb),
			    output_data);
	gtk_box_pack_start (GTK_BOX (hbox), yes_but, TRUE, FALSE, 2);
	gtk_widget_show (yes_but);

	no_but = gtk_button_new_with_label (_("No thanks"));
	gtk_signal_connect (GTK_OBJECT (no_but), "clicked",
			    GTK_SIGNAL_FUNC (no_overwrite_cb),
			    output_data);
	gtk_box_pack_start (GTK_BOX (hbox), no_but, TRUE, FALSE, 2);
	gtk_widget_show (no_but);

	gtk_widget_show (err_win);
}

gint *launch_print_job_cb (GtkWidget *widget,
			   PrintData *print_data)
{
	OutputData *output_data;
	PrintDirection direction;
	gchar *command, *print_command;
	gboolean an_error_occured;

	an_error_occured = FALSE;
	output_data = g_malloc (sizeof (OutputData));
	output_data->document = print_data->document;

	direction = gtk_toggle_button_get_active
		(GTK_TOGGLE_BUTTON (print_data->widgets.printer_rb));

	if (gtk_toggle_button_get_active
	    (GTK_TOGGLE_BUTTON (print_data->widgets.all_pages_rb))) {
		output_data->from_page = 1;
		output_data->to_page =
			print_data->document->nbr_images + 1;
	} else if (gtk_toggle_button_get_active
		   (GTK_TOGGLE_BUTTON (print_data->widgets.from_to_rb))) {
		output_data->from_page =
			gtk_spin_button_get_value_as_int 
			(GTK_SPIN_BUTTON (print_data->widgets.from_sp_but));
		output_data->to_page =
			gtk_spin_button_get_value_as_int
			(GTK_SPIN_BUTTON (print_data->widgets.to_sp_but));
	} else
		output_data->from_page = output_data->to_page =
			print_data->current_page->nbr + 1;

	if (direction) {
		print_command = gtk_entry_get_text
			(GTK_ENTRY (print_data->widgets.printer_cmd_en));
		output_data->output_stream =
			create_temp_print_file (&output_data->out_filename);
	} else {
		output_data->out_filename = g_strdup
			(gtk_entry_get_text (GTK_ENTRY
					     (print_data->widgets.output_file_en)));
		output_data->output_stream = fopen
			(output_data->out_filename, "wx");
	}

	if (output_data->output_stream) {
		output_document (output_data);
		fclose (output_data->output_stream);
		if (direction == PRINT_TO_FILE)
			g_free (output_data->out_filename);
	} else {
		if (direction == PRINT_TO_FILE)
			file_exists_dlg (output_data,
					 widget->parent->parent);
		an_error_occured = TRUE;
	}

	if (direction == PRINT_TO_PRINTER) {
		command = g_strdup_printf ("%s -P%s %s",
					   print_command,
					   print_data->printer,
					   output_data->out_filename);
  		if (system (command))
			an_error_occured = TRUE;
		unlink (output_data->out_filename);
		g_free (command);
	}

	if (!an_error_occured) {
		gtk_widget_destroy (widget->parent->parent);
		g_free (output_data->out_filename);
		g_free (output_data);
	}

	return FALSE;
}

gint *cancel_print_cb (GtkWidget *widget, gpointer null)
{
	gtk_widget_destroy (widget->parent->parent);

	return FALSE;
}

GtkWidget *print_dialog (ViewerData *viewer_data)
{
	GtkWidget *print_dlg, *table,
		*output_frame, *page_frame,
		*print_but, *cancel_but;
	PrintData *print_data;

	print_data = g_malloc (sizeof (PrintData));
	print_data->document = viewer_data->fax_file;
	print_data->current_page = viewer_data->current_page;

	print_dlg = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_transient_for (GTK_WINDOW (print_dlg),
				      GTK_WINDOW
				      (viewer_data->viewer_window));
	gtk_window_set_title (GTK_WINDOW (print_dlg), _("Print..."));
	gtk_window_set_policy (GTK_WINDOW (print_dlg), FALSE, FALSE, TRUE);
	gtk_window_set_modal (GTK_WINDOW (print_dlg), TRUE);
	gtk_container_set_border_width (GTK_CONTAINER (print_dlg), 3);

	table = gtk_table_new (3, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 5);
	gtk_table_set_col_spacings (GTK_TABLE (table), 5);
	gtk_container_add (GTK_CONTAINER (print_dlg), table);
	gtk_widget_show (table);

	output_frame = make_output_frame (print_data);
	gtk_table_attach_defaults (GTK_TABLE (table), output_frame,
				   0, 2, 0, 1); 
	gtk_widget_show (output_frame);

	page_frame = make_page_frame (print_data);
	gtk_table_attach_defaults (GTK_TABLE (table), page_frame,
				   0, 2, 1, 2); 
	gtk_widget_show (page_frame);

	print_but = gtk_button_new_with_label (_("Print"));
	gtk_signal_connect (GTK_OBJECT (print_but), "clicked",
			    GTK_SIGNAL_FUNC (launch_print_job_cb),
			    print_data);
	gtk_table_attach_defaults (GTK_TABLE (table),
				   print_but,
				   0, 1, 2, 3); 
	gtk_widget_show (print_but);

	cancel_but = gtk_button_new_with_label (_("Cancel"));
	gtk_signal_connect (GTK_OBJECT (cancel_but), "clicked",
			    GTK_SIGNAL_FUNC (cancel_print_cb),
			    NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), cancel_but,
				   1, 2, 2, 3); 
	gtk_widget_show (cancel_but);

	gtk_object_set_data_full (GTK_OBJECT (print_dlg),
				  "print_data",
				  print_data,
				  g_free);

	return print_dlg;
}

gint print_cb (void *ignored, ViewerData *viewer_data)
{
	viewer_data->print_dialog =
		print_dialog (viewer_data);
	gtk_widget_show (viewer_data->print_dialog);

	return FALSE;
}

#else  /* ! __UNIX__ */

void windows_print (ViewerData *viewer_data, PRINTDLG *prt_params)
{
	
}

void print_cb (void *ignored, ViewerData *viewer_data)
{
	int result;
	PRINTDLG fax_pdlg;

	memset (&fax_pdlg, 0, sizeof (PRINTDLG));

	fax_pdlg.hwndOwner = NULL;
        fax_pdlg.hDevNames = NULL;
	fax_pdlg.Flags = PD_RETURNDC | PD_PAGENUMS | PD_NOSELECTION;
        fax_pdlg.nMinPage = 1;
        fax_pdlg.nMaxPage = viewer_data->fax_file->nbr_images + 1;
        fax_pdlg.nFromPage = 1;
        fax_pdlg.nToPage = viewer_data->fax_file->nbr_images + 1;
	fax_pdlg.nCopies = 1;
	fax_pdlg.lStructSize = sizeof (PRINTDLG);

	result = PrintDlg (&fax_pdlg);

	if (result)
		windows_print (viewer_data, &fax_pdlg);
}
#endif /* __UNIX__ */
